import subprocess
from rich import inspect, print
from rich.table import Table
from . import tools
from .tools import settings
import time


def execute(command):
    # print(command)
    result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    # inspect(result)
    # exit(1)
    if result.returncode != 0:
        print(f"Command [b]{command}[/] failed:\n {result.stderr}")
        exit(3)

    return result


def stop(session_definition):
    if is_active(session_definition):
        print("Stop {type} forward on [b]{hostname}[/b] {remote_host}:{remote_port}.".format(**session_definition))
        stop_commands = [
            'screen -ls',
            'grep "{name}"',
            'grep -i detached',
            'cut -d. -f1',
            'awk "{{print $1}}"',
            'xargs -I {{}} screen -XS "{{}}" quit'
        ]
        # stop_command = ' | '.join(stop_commands)
        # print(session_definition)
        stop_command = "tmux kill-session -t '{name}'".format(**session_definition)
        execute(stop_command)


def is_active(definition):
    sessions = get_active()

    # print(definition.get('name'))
    for session in sessions:
        shared_items = {k: session[k] for k in session if k in definition and session[k] == definition[k]}
        # print(len(shared_items))
        if len(shared_items) >= 7:
            # print(definition)
            # print(session)
            return True

    return False


def start(session_definition, reconnect: bool = False):
    if reconnect and is_active(session_definition):
        # print(f"[b]Stopping '{session_name}'[/b]")
        stop(session_definition)

    if not is_active(session_definition):
        print("Start {type} forward on [b]{hostname}[/b] {remote_host}:{remote_port}.".format(**session_definition))
        session_definition['shell_command'] = 'ping localhost'

        if session_definition.get('type') == 'remote':
            ssh = 'ssh -o ExitOnForwardFailure=yes -R {local_host}:{local_port}:{remote_host}:{remote_port} {hostname}'
        else:
            ssh = 'ssh -o ExitOnForwardFailure=yes -L {local_host}:{local_port}:{remote_host}:{remote_port} {hostname}'

        session_definition['ssh_command'] = ssh.format(**session_definition)

        # start_command = "screen -dmS '{name}' {ssh_command}  -- {shell_command}"
        start_command = "tmux new-session -d -s '{name}' '{ssh_command} -- {shell_command}'"
        # inspect(session_definition)
        result = execute(start_command.format(**session_definition))
        return result
    else:
        print("Ignoring {type} forward on [b]{hostname}[/b] {remote_host}:{remote_port}.".format(**session_definition))


def get_active():
    command = 'screen -ls | grep -i detached | cut -f2'
    command = "tmux ls | grep 'pfm_session' | cut -d ':' -f 1"
    result = execute(command)
    session_id_list = result.stdout.split("\n")

    sessions = []
    for session_id in session_id_list:
        if len(session_id) == 0:
            continue

        session_data = session_id.replace('_', '.').replace(':', '')
        # print(session_id)
        values = session_data.split('|')

        if len(values) == 8:
            filler, schema, hostname, remote_host, remote_port, local_host, local_port, forward_type = values

            session = {
                'schema': schema,
                'hostname': hostname,
                'type': forward_type,
                'remote_host': remote_host,
                'remote_port': remote_port,
                'local_host': local_host,
                'local_port': local_port,
                'name': session_id
            }

            # print(session)

            definition = get_session_definition(session)
            if definition:
                definition.update(session)
            else:
                definition = session

            sessions.append(definition)

    from operator import itemgetter
    sorted_sessions = sorted(sessions, key=itemgetter('schema'))

    return sorted_sessions


def filter_session(session, schema: str = None, host: str = None, port: str = None):
    if schema and schema not in session.get('schema'):
        return True
    if host and host not in session.get('hostname'):
        return True
    if port and port != session.get('local_port') and port != session.get('remote_port'):
        return True


def list_from_active(key, filter_string: str = ''):
    active_sessions = get_active()
    items = []
    for session in active_sessions:
        item = session.get(key)
        if filter_string not in item:
            continue
        if item not in items:
            items.append(item)

    return items


def prepare_table(show_alias: bool = False, show_link: bool = False, show_schema: bool = True, title: str = 'Schemas'):
    table = Table(show_edge=settings.get('table_border', 'false') == 'true')

    if show_schema:
        table.add_column("Schema", style="green", width=20)

    table.add_column("Hostname", justify="left", style="yellow", width=30)
    table.add_column("Type", justify="right", style="yellow", width=8)
    table.add_column("Local host", justify="right", style="white", width=15)
    table.add_column("Local port", justify="right", style="white", width=7)
    table.add_column("Remote host", justify="right", style="yellow", width=15)
    table.add_column("Remote port", justify="right", style="yellow", width=7)

    if show_alias:
        table.add_column("Alias", justify="right", style="cyan", width=10)

    if show_link:
        table.add_column("URL", style="cyan", width=60)

    return table


def show_active_sessions(schema: str = None, host: str = None, port: int = None):
    sessions = get_active()

    table = prepare_table(True, True)

    count = 0

    sorted_sessions = sorted(sessions, key=lambda d: d['schema'])

    for session in sorted_sessions:
        if filter_session(session, schema, host, port):
            continue
        count += 1

        last_row = False
        if count < len(sorted_sessions):
            if sorted_sessions[count]['schema'] != session['schema']:
                last_row = True

        row = [
            session.get('schema'),
            session.get('hostname'),
            session.get('type'),
            session.get('local_host'),
            session.get('local_port'),
            session.get('remote_host'),
            session.get('remote_port'),
            session.get('alias', '').upper(),
            session.get('link', '').format(**session),
        ]

        table.add_row(*row, end_section=last_row)

    if count > 0:
        print(table)
        if count == 1:
            message = f"[b]{count}[/b] active session.\n\n"
        else:
            message = f"[b]{count}[/b] active sessions.\n\n"
    else:
        message = f"Move along, nothing to see here."

    print(message)


def get_session_definition(session):
    schemas = tools.get_schemas()

    schema = schemas.get(session.get('schema'), [])
    for session_definition in schema:
        if session_definition.get('hostname') != session.get('hostname'):
            continue
        if session_definition.get('remote_port') != session.get('remote_port'):
            continue

        return session_definition




import os
import yaml
from rich import print

settings_file_location = os.path.expanduser('~/.ssh/pfm.settings.yaml')
settings = {
    'show_schema_link': 'false',
    'wait_after_start': 0.5,
    'wait_after_stop': 0.5,
    'table_border': 'false',
    'show_pid': 'false',
    'schemas': {}
}


def load_settings():
    global settings, settings_file_location

    if not os.path.isfile(settings_file_location):
        save_settings()

    with open(settings_file_location, "r") as stream:
        try:
            loaded_settings = yaml.load(stream, Loader=yaml.BaseLoader)
            settings.update(loaded_settings)

            if loaded_settings != settings:
                save_settings()
        except yaml.YAMLError as exc:
            print(exc)


def save_settings():
    global settings, settings_file_location

    print(f"[b]Updating configuration file on '{settings_file_location}'[/b]")

    with open(settings_file_location, "w") as stream:
        try:
            yaml.dump(settings, stream)
        except yaml.YAMLError as exc:
            print(exc)


def session_from_schema(schema_name, forward_definition):
    session_definition = {
        'schema': schema_name,
        'hostname': forward_definition.get('hostname'),
        'type': forward_definition.get('type', 'local'),
        'local_host': forward_definition.get('local_host', '127.0.0.1'),
        'local_port': str(forward_definition.get('local_port', '-')),
        'remote_host': forward_definition.get('remote_host', '127.0.0.1'),
        'remote_port': str(forward_definition.get('remote_port')),
        'alias': forward_definition.get('alias', '').upper(),
        'link': forward_definition.get('link')
    }

    name = forward_definition.get('name')
    if not name:
        name_string = 'pfm_session|{schema}|{hostname}|{remote_host}|{remote_port}|{local_host}|{local_port}|{type}'
        name = name_string.format(**session_definition)

    session_definition['name'] = name

    return session_definition


def get_schemas():
    global settings

    return settings.get('schemas', {})


def get_schema(schema_name):
    schemas = get_schemas()
    schema = schemas.get(schema_name)

    # inspect(schema)
    return schema


def save_forward(forward_definition):
    schemas = get_schemas()

    schema_name = forward_definition['schema']
    if schema_name not in schemas:
        schemas[schema_name] = []

    session_definition = {
        'hostname': forward_definition.get('hostname'),
        'type': forward_definition.get('type'),
        'local_host': forward_definition.get('local_host', '127.0.0.1'),
        'local_port': str(forward_definition.get('local_port', forward_definition.get('remote_port'))),
        'remote_host': forward_definition.get('remote_host', '127.0.0.1'),
        'remote_port': str(forward_definition.get('remote_port')),
    }

    if session_definition not in schemas[schema_name]:
        schemas[schema_name].append(session_definition)
        save_settings()



from . import tools, forward_sessions
from click.shell_completion import CompletionItem


def sc_schemas(ctx, param, incomplete):
    schemas = tools.settings.get('schemas')
    result = []

    for schema_name in schemas:
        if not schema_name.startswith(incomplete):
            continue

        schema = schemas.get(schema_name)
        item = CompletionItem(schema_name, help=f"{len(schema)} session(s) for {schema_name}")
        result.append(item)
    return result


def ac_schemas(incomplete):
    schemas = tools.settings.get('schemas')
    result = []

    for schema_name in schemas:
        if not schema_name.startswith(incomplete):
            continue

        schema = schemas.get(schema_name)
        item = (schema_name, f"{len(schema)} session(s) for {schema_name}")
        result.append(item)
    return result


def sc_active_schemas(ctx, param, incomplete):
    return forward_sessions.list_from_active('schema', incomplete)


def ac_active_schemas(incomplete):
    return forward_sessions.list_from_active('schema', incomplete)


def sc_active_remote_port(ctx, param, incomplete):
    return forward_sessions.list_from_active('remote_port')


def sc_active_hosts(ctx, param, incomplete):
    return forward_sessions.list_from_active('hostname')


def ac_hosts(incomplete):
    from sshconf import read_ssh_config
    from os.path import expanduser
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            result.append(host)
    return result


def sc_hosts(ctx, param, incomplete):
    from sshconf import read_ssh_config
    from os.path import expanduser
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            item = CompletionItem(host)
            result.append(item)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['port_forward_manager']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Pygments>=2.13.0,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'colorama>=0.4.6,<0.5.0',
 'commonmark>=0.9.1,<0.10.0',
 'rich>=12.6.0,<13.0.0',
 'shellingham>=1.5.0,<2.0.0',
 'sshconf>=0.2.5,<0.3.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['pfm = port_forward_manager.cli:run']}

setup_kwargs = {
    'name': 'port-forward-manager',
    'version': '1.5.2',
    'description': 'Port forwarding manager',
    'long_description': '# Port Forward Manager (PFM)\nPFM allows quick and easy management of ports forwarded over SSH (local or remote).\n\n# Installation\n\n```\n# Install python poetry https://python-poetry.org/docs/\nbrew install poetry\nor\ncurl -sSL https://install.python-poetry.org | python3 -\n\n# Install PFM\npip install --upgrade port-forward-manager\n\n\n```\n\nSettings file is stored in ~/.ssh/pfm.settings.yaml\n\n## Configure autocomplete\n\nAdd to end of ~/.zshrc\n\n```\nfpath+=~/.zfunc\n\nautoload -Uz compinit\ncompinit\nzstyle \':completion:*\' menu select\n\n```\n\nGenerate autocomplete configuration\n\n```\npfm --install-completion\nsource .zshrc\n```\n# About\n## Settings\nPFM will automatically generate a default configuration file and update new settings to their default values.\n\n\n### show_schema_link\nToggle the ability to show/hide the schema when showing the list of schemas.\n### wait_after_start \nHow long, in seconds, to wait after starting sessions.\n### table_border\nToggle the table border\n### show_pid\nToggle the screen PID\n\nExample settings file:\n```\nschemas:\n    local_proxy:\n      - hostname: some.proxy.host\n        remote_port: 8888\n        type: local\n    remote-server:\n      - hostname: example.host\n        local_port: 1234\n        remote_port: 8080\n        type: local\n      - hostname: example.host\n        local_port: 8888\n        remote_port: 8888\n        type: local\nshow_pid: \'false\'\nshow_schema_link: \'false\'\ntable_border: \'true\'\nwait_after_start: \'0.5\'\nwait_after_stop: 0.5\n```\n\n## Commands\n### config\nShow active sessions\n### forward\nStart a forwarding session\n### schemas\nList configured schemas\n### shutdown\nStop all active sessions\n### start\nStart a schema of forwarding sessions\n### status\nShow active sessions\n### stop\nStop sessions from active schema, host or port\n###version\nShow installed version\n\n# Development\nSetup development environment\n```\ngit clone git@github.com:kxiros/port-forward-manager.git\ncd port-forward-manager\npoetry shell\npoetry install\n```\n\nBuilding python package\n\n```\npython -m build\n```\n\n\nTo Install development version:\n\n```\npip install -e cloned_directory_path\n```\n\nRelease\n```\n#Example to github\ngh release create 1.3 dist/port-forward-manager-1.3.tar.gz -t "Minor fixes" --generate-notes\n\n#Publish on Pypi\n# Configure pypi token\n# poetry config pypi-token.pypi <token>\npoetry publish\n```',
    'author': 'Vladimiro Casinha',
    'author_email': 'vcasinha@kjoo.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kxiros/port-forward-manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import pkg_resources
import random
from rich import print
import typer
import time

from .forward_sessions import prepare_table, show_active_sessions
from .cli_autocomplete import ac_schemas, ac_hosts
from .cli_autocomplete import sc_schemas, sc_hosts, ac_active_schemas, sc_active_remote_port, sc_active_hosts
from . import tools, forward_sessions


app = typer.Typer(no_args_is_help=True)
tools.load_settings()


@app.command()
def schemas(schema_filter: str = typer.Argument(None, autocompletion=ac_schemas)):
    """
    List configured schemas
    """
    schema_list = list(tools.settings.get('schemas').keys())

    show_link = tools.settings.get('show_schema_link', 'false') == 'true'
    table = prepare_table(True, show_link, True)
    for schema_name in schema_list:
        if schema_filter and schema_filter not in schema_name:
            continue

        schema = tools.get_schema(schema_name)
        count = 0
        for forward_definition in schema:
            count += 1
            session_definition = tools.session_from_schema(schema_name, forward_definition)
            del(session_definition['name'])

            if not show_link:
                del (session_definition['link'])
            # del (session_definition['schema'])
            last_row = False
            if count == len(schema):
                last_row = True

            row = [
                session_definition.get('schema'),
                session_definition.get('hostname'),
                session_definition.get('type'),
                session_definition.get('local_host'),
                session_definition.get('local_port'),
                session_definition.get('remote_host'),
                session_definition.get('remote_port'),
                session_definition.get('alias')
            ]

            table.add_row(*row, end_section=last_row)
        # table.add_row(end_section=True)
    print(table)


def generate_new_port(port_list):
    while True:
        port = str(random.randint(9100, 9500))
        if port not in port_list:
            port_list.append(port)
            return port


@app.command()
def start(schema_name: str = typer.Argument(..., autocompletion=ac_schemas),
          force: bool = typer.Option(None, help="Force sessions reconnection")):
    """
    Start a schema of forwarding sessions
    """
    schema = tools.get_schema(schema_name)
    active_ports = forward_sessions.list_from_active('local_port')
    if schema is None:
        print("[b]Schema '{0}' is unknown[/b]".format(schema_name))
        exit(-1)

    for forward_definition in schema:
        if 'local_port' not in forward_definition:
            new_port = generate_new_port(active_ports)
            forward_definition['local_port'] = new_port

        session_definition = tools.session_from_schema(schema_name, forward_definition)
        # print(session_definition)
        forward_sessions.start(session_definition, force)

    wait_time = float(tools.settings.get('wait_after_start', 0.5))
    time.sleep(wait_time)

    show_active_sessions()


@app.command()
def stop(schema: str = typer.Argument(None, autocompletion=ac_active_schemas),
         hostname: str = typer.Option(None, shell_complete=sc_active_hosts),
         port: str = typer.Option(None, shell_complete=sc_active_remote_port)):
    """
    Stop sessions from active schema, host or port
    """

    if not schema and not hostname and not port:
        print("[b]Pick a schema, host or host and port or --all[/b]")
        exit(-1)

    sessions = forward_sessions.get_active()
    for session in sessions:
        if forward_sessions.filter_session(session, schema, hostname, port):
            continue

        # print(session)
        session_definition = tools.session_from_schema(session['schema'], session)
        forward_sessions.stop(session_definition)

    wait_time = float(tools.settings.get('wait_after_stop', 0.5))
    time.sleep(wait_time)
    show_active_sessions()


@app.command()
def shutdown():
    """
    Stop all active sessions
    """

    sessions = forward_sessions.get_active()
    for session in sessions:
        session_definition = tools.session_from_schema(session['schema'], session)
        forward_sessions.stop(session_definition)

    show_active_sessions()


@app.command()
def forward(hostname: str = typer.Argument(..., autocompletion=ac_hosts),
            remote_port: int = typer.Argument(...),
            local_port: int = None,
            remote_host: str = '127.0.0.1',
            local_host: str = '127.0.0.1',
            remote: bool = False,
            schema: str = typer.Option('user-cli', shell_complete=ac_schemas), save: bool = False):
    """
    Start a forwarding session
    """

    if save and schema == 'user-cli':
        print("[b]To save, schema must be different from default 'user-cli'[/b]")
        exit(1)

    session_definition = {
        'hostname': hostname,
        'local_port': local_port or remote_port,
        'local_host': local_host,
        'remote_port': remote_port,
        'remote_host': remote_host,
        'type': 'remote' if remote else 'local'
    }

    forward_definition = tools.session_from_schema(schema, session_definition)
    forward_sessions.start(forward_definition)

    wait_time = float(tools.settings.get('wait_after_start', 0.5))
    time.sleep(wait_time)

    if save:
        tools.save_forward(forward_definition)

    show_active_sessions()


@app.command()
def status(schema: str = typer.Option(None, shell_complete=sc_schemas),
           host: str = typer.Option(None, shell_complete=sc_hosts),
           port: int = None):
    """
    Show active sessions
    """
    show_active_sessions(schema, host, port)


@app.command()
def config():
    """
    Show active sessions
    """

    print(tools.settings)


@app.command()
def version():
    """
    Show active sessions
    """

    current_version = pkg_resources.get_distribution("port-forward-manager").version
    print(f"Port Forward Manager [bold white]v{current_version}[/]")


@app.command()
def state():
    result = forward_sessions.list_from_active('local_port')
    print(result)
def run():
    app()


if __name__ == "__main__":
    app()
